# coding: utf-8
# Date      : 2019/4/23
# Author    : kylin
# PROJECT   : credit
# File      : aio_downloader
import traceback
from aiohttp import ClientSession
from aiocrawler.response import Response
from aiocrawler.downloaders.downloader import BaseDownloader
from aiocrawler.settings import BaseSettings


class AioDownloader(BaseDownloader):
    def __init__(self, settings: BaseSettings):
        BaseDownloader.__init__(self, settings)

    async def get_response(self, request):
        session = ClientSession()
        try:
            if request['cookies']:
                session.cookie_jar.update_cookies(request['cookies'])

            if request['method'] == 'GET':
                resp = await session.get(request['url'],
                                         params=request['params'],
                                         proxy=request['proxy'],
                                         headers=request['headers'],
                                         timeout=request['timeout'],
                                         )
            else:
                resp = await session.post(url=request['url'],
                                          data=request['params'],
                                          proxy=request['proxy'],
                                          headers=request['headers'],
                                          timeout=request['timeout'],
                                          )

            status = resp.status
            text = await resp.text(encoding=request['encoding'])
            cookies = resp.cookies

            response = Response(text=text, status=status, cookies=cookies)
            if request['cookies']:
                response.cookies.update(request['cookies'])

            return response

        except Exception as e:
            self.logger.error(traceback.format_exc(limit=10))
            return e

        finally:
            await session.close()




# coding: utf-8
# Date      : 2019/4/25
# Author    : kylin
# PROJECT   : aiocrawler
# File      : allowed_code_middleware
from aiocrawler.settings import BaseSettings
from aiocrawler.request import Request
from aiocrawler.response import Response
from aiocrawler.middlewares.middleware import BaseDownloaderMiddleware


class AllowedCodesMiddleware(BaseDownloaderMiddleware):
    def __init__(self, settings: BaseSettings):
        BaseDownloaderMiddleware.__init__(self, settings)

    def process_response(self, request: Request, response: Response):
        if not 200 <= response.status <= 301 and response.status not in self.settings.ALLOWED_CODES:
            self.logger.debug('The response status <{code}> is not in ALLOWED_CODES', code=response.status)

            retry_count = request['meta'].get('retry_count', 0)
            retry_count += 1

            request['meta']['retry_count'] = retry_count
            return request



# coding: utf-8
# Date      : 2019/4/23
# Author    : kylin
# PROJECT   : aiocrawler
# File      : set_default_middleware
from aiocrawler.settings import BaseSettings
from aiocrawler.request import Request
from aiocrawler.middlewares.middleware import BaseDownloaderMiddleware


class SetDefaultRequestMiddleware(BaseDownloaderMiddleware):
    def __init__(self, settings: BaseSettings):
        BaseDownloaderMiddleware.__init__(self, settings)

    def process_request(self, request: Request):
        if request['timeout'] is None:
            request['timeout'] = self.settings.DEFAULT_TIMEOUT
        if request['headers'] is None:
            request['headers'] = self.settings.DEFAULT_HEADERS

        if request['meta'] is None:
            request['meta'] = {}



# coding: utf-8
# Date      : 2019/4/23
# Author    : kylin1020
# PROJECT   : aiocrawler
# File      : redis_filter
from re import findall
from hashlib import sha1
from yarl import URL
from aiocrawler.item import Item
from aiocrawler.filters.filter import BaseFilter
from aiocrawler.settings import BaseSettings
from aiocrawler.request import Request
from aioredis import create_pool, ConnectionsPool


class RedisFilter(BaseFilter):

    def __init__(self, settings: BaseSettings):
        BaseFilter.__init__(self, settings)

        self.__redis_pool: ConnectionsPool = None
        self.__redis_filters_key = settings.REDIS_PROJECT_NAME + ':filters'

    async def __initialize_redis_pool(self):
        if not self.settings.REDIS_URL:
            raise ValueError('REDIS_URL is not configured in {setting_name}'.format(
                setting_name=self.settings.__class__.__name__))
        else:
            self.__redis_pool = await create_pool(self.settings.REDIS_URL)

    async def filter_request(self, request: Request):
        if not self.__redis_pool:
            await self.__initialize_redis_pool()

        if request['dont_filter']:
            return request

        elif await self.exist_request(request):
            return None
        return request

    async def filter_item(self, item: Item):
        if not self.__redis_pool:
            await self.__initialize_redis_pool()

        hex_data = self.sha_item(item)
        resp = await self.__redis_pool.execute('sismember', self.__redis_filters_key, hex_data)
        if resp:
            self.logger.debug('The Item has existed in The Redis Server: {}'.format(hex_data))
            return item
        else:
            await self.__redis_pool.execute('sadd', self.__redis_filters_key, hex_data)
            return None

    async def exist_request(self, request: Request):

        hex_data = self.sha_request(request)
        resp = await self.__redis_pool.execute('sismember', self.__redis_filters_key, hex_data)
        if resp:
            self.logger.debug('The Request has existed in The Redis Server: {}'.format(hex_data))
            return True
        else:
            await self.__redis_pool.execute('sadd', self.__redis_filters_key, hex_data)
            return False

    @staticmethod
    def sha_item(item: Item):
        sha = sha1()
        for key, value in item.items():
            sha.update(str(key).encode())
            sha.update(str(value).encode())
        return sha.hexdigest()

    def sha_request(self, request: Request):
        sha = sha1()

        base_url, params = self.parse_url(request['url'])

        if request['method'] == 'GET':
            if request['params']:
                for data in request['params'].items():
                    params.append(data)

            params = sorted(params, key=lambda x: x)
        else:
            for key, value in sorted(request['params'].items(), key=lambda x: x[0]):
                sha.update(str(key).encode())
                sha.update(str(value).encode())

        for key, value in params:
            sha.update(str(key).encode())
            sha.update(str(value).encode())

        sha.update(request['method'].encode())
        sha.update(request['url'].encode())

        return sha.hexdigest()

    @staticmethod
    def parse_url(url: str):
        url = URL(url)

        base_url = str(url.parent) + url.path

        param_pattern = r'(\w+)=([^=&]*)'
        params = findall(param_pattern, url.query_string)
        return base_url, params



# coding: utf-8

from aiocrawler.request import Request
from aiocrawler.item import Item
from aiocrawler.settings import BaseSettings


class BaseScheduler(object):
    def __init__(self, settings: BaseSettings):
        self.settings = settings
        self.logger = self.settings.LOGGER

    async def get_word(self):
        raise NotImplementedError('{} get_word not define'.format(self.__class__.__name__))

    async def get_request(self):
        raise NotImplementedError('{} get_request not define'.format(self.__class__.__name__))

    async def send_request(self, request: Request):
        raise NotImplementedError('{} send_request not define'.format(self.__class__.__name__))

    async def send_item(self, item: Item):
        raise NotImplementedError('{} send_item not define'.format(self.__class__.__name__))

    async def get_total_request(self):
        raise NotImplementedError('{} get_total_request not define'.format(self.__class__.__name__))



# coding: utf-8
# Date      : 2019/4/23
# Author    : kylin
# PROJECT   : aiocrawler
# File      : user_agent_middleware
from aiocrawler.middlewares.middleware import BaseDownloaderMiddleware
from aiocrawler.settings import BaseSettings
from aiocrawler.request import Request


class UserAgentMiddleware(BaseDownloaderMiddleware):
    def __init__(self, settings: BaseSettings):
        BaseDownloaderMiddleware.__init__(self, settings)
        from fake_useragent import UserAgent
        self.__ua = UserAgent()

    def process_request(self, request: Request):
        request['headers']['User-Agent'] = self.__ua.random



# coding: utf-8
import sys
from loguru import logger


class BaseSettings:
    logger.remove()
    fmt = "<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "\
          "<level>{level}</level> | "\
          "<cyan>{name}</cyan> <line {line}>: <level>{message}</level>"
    logger.add(sys.stdout, format=fmt)
    logger.add('log/aio-crawler.log', format=fmt, rotation='10 MB')
    LOGGER = logger

    REDIS_URL = None
    REDIS_PROJECT_NAME = None

    MYSQL_HOST = 'localhost'
    MYSQL_USER = 'root'
    MYSQL_PASSWORD = None
    MYSQL_DB = None

    CONCURRENT_REQUESTS = 16
    CONCURRENT_WORDS = 16
    DEFAULT_TIMEOUT = 20
    DEFAULT_HEADERS = {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
        'Accept-Language': 'en',
    }

    DOWNLOAD_DALEY = 0
    PROCESS_DALEY = 0.01

    ALLOWED_CODES = []

    DOWNLOADER_MIDDLEWARES = []



# coding: utf-8


class Request(dict):
    def __init__(self, url: str,
                 method: str = 'GET',
                 params: dict = None,
                 dont_filter: bool = False,
                 proxy: str = None,
                 err_callback=None,
                 encoding: str = None,
                 headers: dict = None,
                 format_type: str = 'selector',
                 timeout: int = None,
                 meta: dict = None,
                 cookies: dict = None,
                 callback=None):
        dict.__init__(self)

        self['url'] = url
        self['method'] = method if method in ['GET', 'POST'] else 'GET'
        self['proxy'] = proxy
        self['dont_filter'] = dont_filter
        self['params'] = params

        self['callback'] = callback.__name__ if callback else 'parse'
        self['err_callback'] = err_callback.__name__ if err_callback else 'handle_error'

        self['timeout'] = timeout
        self['headers'] = headers

        self['encoding'] = encoding
        self['cookies'] = cookies
        self['meta'] = meta

        if format_type in ['text', 'json', 'selector']:
            self['handle_way'] = format_type
        else:
            self['handle_way'] = 'selector'



# coding: utf-8
from typing import Union, List
from json import loads
from parsel import Selector
from aiocrawler import BaseSettings
from aiocrawler import Request
from aiocrawler import Response


class Spider(object):
    name: str = None
    words: List[str] = None

    def __init__(self, settings: BaseSettings):
        self.setting = settings
        self.logger = self.setting.LOGGER

    def make_request(self, word: str) -> Union[List[Request], Request]:
        raise NotImplementedError('{} make_request method is not define'.format(self.__class__.__name__))

    def parse(self, response: Response):
        pass

    def __handle__(self, request: Request, response: Response) -> Response:
        if request['handle_way'] == 'json':
            try:
                response.json = loads(response.text)
            except Exception as e:
                self.logger.error(e)
        elif request['handle_way'] == 'selector':
            response.selector = Selector(text=response.text)

        return response

    def handle_error(self, request: Request, exception: Exception) -> Union[Request, None]:
        pass



# coding: utf-8
from aiocrawler.request import Request
from typing import Union
from aiocrawler.settings import BaseSettings
from aiocrawler.item import Item


class BaseFilter(object):
    def __init__(self, settings: BaseSettings):
        self.settings = settings
        self.logger = settings.LOGGER

    async def filter_request(self, request: Request) -> Union[None, Request]:
        raise NotImplementedError('{} filter_request is not define'.format(self.__class__.__name__))

    async def filter_item(self, item: Item):
        raise NotImplementedError('{} filter_item is not define'.format(self.__class__.__name__))



# coding: utf-8
# Date      : 2019/4/23
# Author    : kylin
# PROJECT   : aiocrawler
# File      : redis_to_mysql
import asyncio
import aioredis
import aiomysql
import traceback
import pickle
import json
from numpy import ceil
from aiocrawler.settings import BaseSettings
from aiocrawler.extensions.exporters.exporter import BaseExporter
from aiocrawler.item import Item
from typing import List, Iterable, Union


class RedisToMysql(BaseExporter):
    def __init__(self, settings: BaseSettings, item_class: Item, max_length: int = 128):

        BaseExporter.__init__(self, settings, item_class)

        self.__loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

        self.__redis_pool: aioredis.ConnectionsPool = None
        self.__mysql_pool: aiomysql.Pool = None
        self.__redis_items_key = self.settings.REDIS_PROJECT_NAME + ':items'

        self.__max_length = max_length
        self.__table_name = item_class.__class__.__name__.lower()
        self.__fields = list(self.get_fields(item_class))

    async def __initialize__(self):
        """
        Initialize all necessary components.
        """

        if not self.settings.REDIS_URL or not self.settings.MYSQL_DB:
            raise ValueError('REDIS_URL or MYSQL_DB are not configured in {setting_name}'.format(
                setting_name=self.settings.__class__.__name__))
        else:
            self.logger.debug('Connecting to The Redis Server...')
            self.__redis_pool = await aioredis.create_pool(self.settings.REDIS_URL, loop=self.__loop)

            self.logger.debug('Connecting to The Mysql Server...')
            self.__mysql_pool: aiomysql.Pool = await aiomysql.create_pool(
                host=self.settings.MYSQL_HOST,
                user=self.settings.MYSQL_USER,
                password=self.settings.MYSQL_PASSWORD,
                db=self.settings.MYSQL_DB,
                loop=self.__loop
            )

            async with self.__mysql_pool.acquire() as conn:
                async with conn.cursor() as cur:
                    await self.__create_tables__(cur)

    async def __create_tables__(self, cur):
        await self.__create__(self.item_class, cur)

    async def __create__(self, item_class: Item, cur):
        self.logger.debug(f'Creating the table <name: {self.__table_name}>...')

        columns = ', '.join(map(lambda x: x + f' VARCHAR({self.__max_length})', self.get_fields(item_class)))
        sql = f'CREATE TABLE IF NOT EXISTS {self.__table_name} ' \
            f'({self.__table_name}_id INT NOT NULL AUTO_INCREMENT, {columns}, PRIMARY KEY({self.__table_name}_id));'
        await cur.execute(sql)

    async def __insert_into_mysql(self, item_values: Iterable, cur):

        column_str = ', '.join(self.__fields)

        values_str = ', '.join(['?' for _ in range(len(list(self.__fields)))])
        sql = f'INSERT INTO {self.__table_name} ({column_str}, VALUES ({values_str}));'
        await cur.executemany(sql, item_values)

    async def __get_redis_items(self, start: int, end: int):
        items = await self.__redis_pool.execute('lrange', self.__redis_items_key, start, end)
        count = len(items)
        self.logger.debug(f'Got {count} items <{start}, {end}> from '
                          f'The Redis Server <redis key: {self.__redis_items_key}>')
        return items

    async def main(self):
        total_count = await self.__redis_pool.execute('llen', self.__redis_items_key)
        self.logger.debug(f'Total {total_count} items in The Redis Server <redis key: {self.__redis_items_key}>')

        step = 1000
        batch = int(ceil(total_count / step))
        async with self.__mysql_pool.acquire() as conn:
            async with conn.cursor() as cur:
                for i in range(batch):
                    await self.__redis_to_mysql__(i*step, (i + 1)*step, cur)

    async def __redis_to_mysql__(self, start: int, end: int, cur):
        # sleep(0.05)
        items = await self.__get_redis_items(start, end)
        # items = filter(lambda x: x.__class__.__name__ == self.item_class.__class__.__name__, items)

        await self.__insert_into_mysql(self.__gen_item_data(items), cur)
        self.logger.success(f'The items <{start}, {end}> have been inserted into '
                            f'The Mysql Server <table_name: {self.__table_name}>')

    def __gen_item_data(self, items: Iterable[Item]) -> List:
        fields = self.get_fields(self.item_class)

        for item in items:
            try:
                item = pickle.loads(item)
            except:
                item = json.loads(item)
            data = []
            for column in fields:
                data.append(item.get(column, None))
            yield data

    def run(self):
        try:
            self.__loop.run_until_complete(self.__initialize__())
            self.__loop.run_until_complete(self.main())
        except Exception as e:
            self.logger.error(e)
            self.logger.error(traceback.format_exc(limit=10))
        finally:
            self.__loop.close()



# coding: utf-8
from aiocrawler.request import Request
from aiocrawler.settings import BaseSettings
from aiocrawler.response import Response
from typing import Union


class BaseDownloaderMiddleware(object):
    def __init__(self, settings: BaseSettings):
        self.settings = settings
        self.logger = self.settings.LOGGER

    def process_request(self, request: Request):
        pass

    def process_response(self, request: Request, response: Response) -> Union[None, Request, Response]:
        pass

    def process_exception(self, request: Request, exception: Exception) -> Union[None, Request]:
        pass



# coding: utf-8
# Date      : 2019/4/26
# Author    : kylin
# PROJECT   : aiocrawler
# File      : aiocrawler
import sys
import argparse
from pathlib import Path
from pyclbr import readmodule
from importlib import import_module
from aiocrawler.settings import BaseSettings
from aiocrawler.extensions.templates import SpiderTemplate

logger = BaseSettings.LOGGER


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("commands", choices=["startproject", "run"], help="The Aiocrawler Commands")
    parser.add_argument('name', help="The Project Name you want to start")
    args = parser.parse_args()

    if args.commands == "startproject":
        tmpl = SpiderTemplate(args.name)
        tmpl.gen_project()
    elif args.commands == "run":
        spider = get_module(args.name)
        if not spider:
            logger.error('The Spider name you provided is not found in this directory.')
            return

        try:
            run_module = import_module('run')
            run_module.run(spider)
        except Exception as e:
            logger.error(e)


def get_module(name: str, spider_module_name: str = 'spiders'):
    module = None
    try:
        current_dir = str(Path('').cwd())
        if current_dir not in sys.path:
            sys.path.append(current_dir)

        spider_module = import_module(spider_module_name)

        for module_name in readmodule(spider_module_name).keys():
            module = getattr(spider_module, module_name)
            if name == vars(module).get(name, ''):
                return module
    finally:
        return module



# coding: utf-8
from aiocrawler.settings import BaseSettings
from typing import Union
from aiocrawler.request import Request
from aiocrawler.response import Response


class BaseDownloader(object):
    def __init__(self, settings: BaseSettings):
        self.settings = settings
        self.logger = settings.LOGGER

    async def get_response(self, request: Request) -> Union[Response, Exception, None]:
        raise NotImplementedError('{} get_response is not define'.format(self.__class__.__name__))






# coding: utf-8
# Date      : 2019/4/23
# Author    : kylin1020
# PROJECT   : aiocrawler
# File      : redis_scheduler
from aiocrawler.schedulers.scheduler import BaseScheduler
from aiocrawler.request import Request
from aiocrawler.item import Item
from aiocrawler.settings import BaseSettings
from aioredis import create_pool, ConnectionsPool
import pickle


class RedisScheduler(BaseScheduler):
    def __init__(self, settings: BaseSettings):
        BaseScheduler.__init__(self, settings)

        self.redis_pool: ConnectionsPool = None
        self.__redis_words_key = self.settings.REDIS_PROJECT_NAME + ':words'
        self.__redis_requests_key = self.settings.REDIS_PROJECT_NAME + ':requests'
        self.__redis_items_key = self.settings.REDIS_PROJECT_NAME + ':items'

    async def __initialize_redis_pool(self):
        if not self.settings.REDIS_URL:
            raise ValueError('REDIS_URL are not configured in {setting_name}'.format(
                setting_name=self.settings.__class__.__name__))
        else:
            self.redis_pool = await create_pool(self.settings.REDIS_URL)

    async def get_request(self):
        if not self.redis_pool:
            await self.__initialize_redis_pool()

        request = await self.redis_pool.execute('lpop', self.__redis_requests_key)
        if request:
            request = pickle.loads(request)
        return request

    async def get_word(self):
        if not self.redis_pool:
            await self.__initialize_redis_pool()

        key = await self.redis_pool.execute('lpop', self.__redis_words_key)
        if key:
            key = key.decode()
        return key

    async def send_request(self, request: Request):
        if not self.redis_pool:
            await self.__initialize_redis_pool()

        resp = await self.redis_pool.execute('lpush', self.__redis_requests_key,
                                             pickle.dumps(request))
        if resp:
            self.logger.success('Sent <Request> <url: {url}> to redis server', url=request['url'])
        else:
            self.logger.error('Send <Request> Failed <url: {url}> to redis server', url=request['url'])

    async def send_item(self, item: Item):
        if not self.redis_pool:
            await self.__initialize_redis_pool()

        await self.redis_pool.execute('lpush', self.__redis_items_key, pickle.dumps(item))

    async def get_total_request(self):
        if not self.redis_pool:
            await self.__initialize_redis_pool()

        count = await self.redis_pool.execute('llen', self.__redis_requests_key)
        return count



# coding: utf-8
from pathlib import Path
from datetime import date
from typing import Union
from string import Template
import traceback
from aiocrawler.settings import BaseSettings


class SpiderTemplate(object):
    def __init__(self, project_name: str, output_dir: Union[Path, str] = ''):
        self.__templates_dir = Path(__file__).parent.parent/'templates'
        self.__logger = BaseSettings.LOGGER

        self.__project_name = project_name
        self.__output_dir = Path(output_dir)
        self.__project_dir = self.__output_dir/self.__project_name

    def gen_project(self):
        if self.__project_dir.is_dir():
            self.__logger.error('The target Project "{project_name}" already exists',
                                project_name=self.__project_name)
            return

        try:
            self.__project_dir.mkdir()
            self.__gen_spiders__()
            self.__gen_settings__()
            self.__gen_middlewares__()
            self.__gen_items__()
            self.__gen_run__()

            self.__logger.success('The Project "{project_name}" was created successfully.',
                                  project_name=self.__project_name)
        except Exception as e:
            files = self.__project_dir.glob('*')
            for file in files:
                file.unlink()
            self.__project_dir.rmdir()

            self.__logger.error(traceback.format_exc(limit=10))

    def gen(self, tmpl_name: str, sub_data: dict):
        tmpl_file = self.__templates_dir/tmpl_name
        output_name = Path(tmpl_name).stem + '.py'

        default_data = {
            'date': date.today(),
            'project_name': self.__project_name,
            'filename': output_name,
        }
        sub_data.update(default_data)

        if tmpl_file.is_file():
            with tmpl_file.open('r') as f:
                content = f.read()
            tmpl = Template(content).substitute(sub_data)

            output = Path(self.__project_dir)/output_name
            with output.open('w') as f:
                f.write(tmpl)

            self.__logger.success('Generated {filename}', filename=str(output))
        else:
            self.__logger.error('{template_name} does not exist', template_name=tmpl_name)

    def __gen_spiders__(self):
        tmpl_name = 'spiders.tmpl'
        sub_data = {
            'classname': self.__project_name.capitalize() + 'Spider',
            'name': self.__project_name.lower()
        }
        self.gen(tmpl_name=tmpl_name, sub_data=sub_data)

    def __gen_run__(self):
        tmpl_name = 'run.tmpl'
        sub_data = {
            'settings_name': self.__project_name.capitalize() + 'Settings',
        }
        self.gen(tmpl_name=tmpl_name, sub_data=sub_data)

    def __gen_middlewares__(self):
        tmpl_name = 'middlewares.tmpl'
        sub_data = {
            'classname': self.__project_name.capitalize() + 'Middleware'
        }
        self.gen(tmpl_name=tmpl_name, sub_data=sub_data)

    def __gen_settings__(self):
        tmpl_name = 'settings.tmpl'
        sub_data = {
            'classname': self.__project_name.capitalize() + 'Settings',
            'middleware_name': self.__project_name.capitalize() + 'Middleware'
        }
        self.gen(tmpl_name=tmpl_name, sub_data=sub_data)

    def __gen_items__(self):
        tmpl_name = 'items.tmpl'
        sub_data = {
            'classname': self.__project_name.capitalize() + 'Item'
        }
        self.gen(tmpl_name=tmpl_name, sub_data=sub_data)



# coding: utf-8
import traceback
import asyncio
import signal
from asyncio import AbstractEventLoop
from time import sleep
from aiocrawler.item import Item
from aiocrawler.field import Field
from aiocrawler.request import Request
from aiocrawler.filters.filter import BaseFilter
from aiocrawler.settings import BaseSettings
from aiocrawler.schedulers.scheduler import BaseScheduler
from typing import List, Union, Iterator, Tuple
from aiocrawler.middlewares.middleware import BaseDownloaderMiddleware
from aiocrawler.middlewares.user_agent_middleware import UserAgentMiddleware
from aiocrawler.middlewares.set_default_middleware import SetDefaultRequestMiddleware
from aiocrawler.middlewares.allowed_codes_middleware import AllowedCodesMiddleware
from aiocrawler.response import Response
from aiocrawler.spider import Spider
from aiocrawler.downloaders.downloader import BaseDownloader


class Engine(object):
    """
    The Engine schedules all components.
    """
    def __init__(self, spider: Spider,
                 settings: BaseSettings = None,
                 downloader_middlewares: List[Tuple[BaseDownloaderMiddleware, int]] = None,
                 filters: BaseFilter = None,
                 scheduler: BaseScheduler = None,
                 downloader: BaseDownloader = None
                 ):

        self.__spider = spider
        self.__scheduler = scheduler
        self.__settings = settings

        self.__filters = filters
        self.__downloader_middlewares = downloader_middlewares

        self.__downloader: BaseDownloader = downloader
        self.__logger = settings.LOGGER

        self.__crawled_count__ = 0
        self.__item_count__ = 0
        self.__log_interval__ = 30
        self.__loop: AbstractEventLoop = None

        self.__signal_int_count = 0
        self.__is_stop = False

    async def __process_run(self):
        """
        Initialize all necessary components.
        """
        if not self.__settings:
            self.__settings = BaseSettings()

        if not self.__downloader:
            from aiocrawler.downloaders.aio_downloader import AioDownloader
            self.__downloader = AioDownloader(self.__settings)

        if not self.__scheduler:
            from aiocrawler.schedulers.redis_scheduler import RedisScheduler
            self.__scheduler = RedisScheduler(settings=self.__settings)

        if not self.__filters:
            # Use Redis Filters by default.
            from aiocrawler.filters.redis_filter import RedisFilter
            self.__filters = RedisFilter(settings=self.__settings)

        if not self.__downloader_middlewares:
            self.__downloader_middlewares = []

        default_middlewares = [
            (SetDefaultRequestMiddleware(self.__settings), 1),
            (UserAgentMiddleware(self.__settings), 2),
            (AllowedCodesMiddleware(self.__settings), 3),
        ]
        self.__downloader_middlewares.extend(default_middlewares)
        self.__downloader_middlewares = sorted(self.__downloader_middlewares, key=lambda x: x[1])

        self.__is_process_run = True

    async def handle_response(self, request: Request, data: Union[Response, Exception, None]):
        """
        Handle the information returned by the downloader.
        :param request: Request
        :param data: Response or Exception
        """
        handled_data = None

        if isinstance(data, Exception):

            for downloader_middleware, _ in self.__downloader_middlewares:
                handled_data = downloader_middleware.process_exception(request, data)
                if handled_data:
                    break

            if handled_data is None:
                err_callback = getattr(self.__spider, request['err_callback'])
                handled_data = err_callback(request, data)

        elif isinstance(data, Response):
            response = self.__spider.__handle__(request, data)
            for downloader_middleware, _ in self.__downloader_middlewares:
                handled_data = downloader_middleware.process_response(request, response)
                if handled_data:
                    if isinstance(handled_data, Response):
                        response = handled_data
                    break

            if isinstance(handled_data, Response) or not handled_data:
                self.__crawled_count__ += 1
                self.__logger.success('Crawled ({status}) <{method} {url}>',
                                      status=response.status,
                                      method=request['method'],
                                      url=request['url']
                                      )

                response.meta = request['meta']
                callback = getattr(self.__spider, request['callback'])
                handled_data = callback(response)

        if not handled_data:
            return

        if not isinstance(handled_data, Iterator) and not isinstance(handled_data, List):
            handled_data = [handled_data]

        tasks = []
        for one in handled_data:
            if isinstance(one, Request):
                tasks.append(asyncio.ensure_future(self.__scheduler.send_request(one)))
            elif isinstance(one, Item):
                self.__item_count__ += 1

                item_copy = one.__class__()
                for field in self.get_fields(one):
                    item_copy[field] = one.get(field, None)

                self.__logger.success('Crawled from <{method} {url}> \n {item}',
                                      method=request['method'], url=request['url'], item=item_copy)
                tasks.append(asyncio.ensure_future(self.__item_filter_and_send__(item_copy)))

        if len(tasks):
            await asyncio.wait(tasks)

    async def __item_filter_and_send__(self, item: Item):
        item = await self.__filters.filter_item(item)
        if item:
            await self.__scheduler.send_item(item)

    @staticmethod
    def get_fields(item: Item):
        for field_name in item.__class__.__dict__:
            if isinstance(getattr(item.__class__, field_name), Field):
                yield field_name

    async def handle_word(self):
        """
        Handle the word from the scheduler.
        """
        try:
            while True:

                sleep(self.__settings.PROCESS_DALEY)
                word = await self.__scheduler.get_word()
                if word:
                    self.__logger.debug('Making Request from word <word: {word}>'.format(word=word))
                    request = self.__spider.make_request(word)
                    if request:
                        await self.__scheduler.send_request(request)

        except asyncio.CancelledError:
            pass
        except Exception as e:
            self.__logger.error(e)
            self.__logger.error(traceback.format_exc(limit=10))

    async def handle_request(self):
        """
        Handle the request from scheduler.
        """
        try:
            while True:

                sleep(self.__settings.PROCESS_DALEY)
                request = await self.__scheduler.get_request()
                if request:
                    self.__logger.debug('Received 1 <Request> <url: {url}> from {scheduler_name}',
                                        url=request['url'],
                                        scheduler_name=self.__scheduler.__class__.__name__)
                    request = await self.__filters.filter_request(request)
                    if request:
                        for downloader_middleware, _ in self.__downloader_middlewares:
                            downloader_middleware.process_request(request)

                        sleep(self.__settings.DOWNLOAD_DALEY)
                        data = await self.__downloader.get_response(request)
                        await self.handle_response(request, data)

        except asyncio.CancelledError:
            pass
        except Exception as e:
            self.__logger.error(e)
            self.__logger.error(traceback.format_exc(limit=10))

    async def __log__(self):
        """
        Log crawled information.
        """
        while True:

            request_count = await self.__scheduler.get_total_request()
            self.__logger.debug('Total Crawled {crawled_count} Pages, {item_count} Items; '
                                'Total {request_count} Requests in The {scheduler_name}',
                                crawled_count=self.__crawled_count__,
                                item_count=self.__item_count__,
                                request_count=request_count,
                                scheduler_name=self.__scheduler.__class__.__name__)
            await asyncio.sleep(self.__log_interval__)

    def signal_int(self, signum, frame):
        self.__signal_int_count += 1
        self.__logger.debug('Received SIGNAL INT {times} times.', times=self.__signal_int_count)
        if self.__signal_int_count >= 2:
            self.__logger.debug('Received SIGNAL INT over 2 times, closing the aiocrawler by force...')
            tasks = asyncio.Task.all_tasks()
            for task in tasks:
                task.cancel()
            asyncio.ensure_future(self.stop_loop())

    async def stop_loop(self):
        if self.__loop:
            self.__loop.stop()

    def run(self):
        """
        Start event loop.
        """
        signal.signal(signal.SIGINT, self.signal_int)

        self.__loop = asyncio.get_event_loop()
        try:
            self.__logger.debug('Initializing The Crawler...')
            self.__loop.run_until_complete(self.__process_run())

            for _ in range(self.__settings.CONCURRENT_WORDS):
                asyncio.ensure_future(self.handle_word())

            for _ in range(self.__settings.CONCURRENT_REQUESTS):
                asyncio.ensure_future(self.handle_request())

            asyncio.ensure_future(self.__log__())
            self.__logger.debug('The Crawler Initialized')

            self.__loop.run_forever()
        finally:
            self.__loop.close()



# coding: utf-8
# Date      : 2019/4/26
# Author    : kylin
# PROJECT   : aiocrawler
# File      : setup
from setuptools import setup, find_packages

setup(
    name="aiocrawler",
    version="1.17",
    keywords=["spider", "asynchronous", "crawler", "distributed"],
    author="kylin1020",
    author_email="kylin0521@gmail.com",
    description="The aiocrawler is a asynchronous/distributed web crawler/spider",
    url="https://github.com/kylin1020/aiocrawler",
    license="MIT",
    packages=find_packages(),
    package_data={"": ["*.py", "*.tmpl"]},
    install_requires=["aiohttp", "aioredis", "yarl", "parsel", "loguru"],
    entry_points={"console_scripts": [
        "aiocrawler = aiocrawler.commands:main"
    ]},
    classifiers=[
        "Environment :: Web Environment",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ]
)



from aiocrawler.request import Request
from aiocrawler.item import Item
from aiocrawler.settings import BaseSettings
from aiocrawler.response import Response
from aiocrawler.field import Field
from aiocrawler.spider import Spider
from aiocrawler.engine import Engine
from aiocrawler.middlewares.middleware import BaseDownloaderMiddleware
from aiocrawler.schedulers.scheduler import BaseScheduler
from aiocrawler.filters.filter import BaseFilter
from aiocrawler.downloaders.downloader import BaseDownloader

__all__ = ['Request',
           'Item',
           'BaseSettings',
           'Field',
           'Response',
           'Spider',
           'BaseDownloaderMiddleware',
           'Engine',
           'BaseScheduler',
           'BaseFilter',
           'BaseDownloader'
           ]



# coding: utf-8
# Date      : 2019/4/23
# Author    : kylin
# PROJECT   : aiocrawler
# File      : exporters
from aiocrawler.settings import BaseSettings
from typing import List, Union
from aiocrawler.item import Item
from aiocrawler.field import Field
from numpy import ceil


class BaseExporter(object):
    def __init__(self, settings: BaseSettings, item_class: Item):
        self.settings = settings
        self.logger = settings.LOGGER
        self.item_class = item_class

    @staticmethod
    def chunk(items: List[Item], batch_size: int = 512):
        batch = int(ceil(len(items) / batch_size))
        for i in range(batch):
            yield items[i*batch_size:(i+1)*batch_size]

    @staticmethod
    def get_fields(item: Item):
        for field_name in item.__class__.__dict__:
            if isinstance(getattr(item.__class__, field_name), Field):
                yield field_name



# coding: utf-8
from http.cookies import SimpleCookie
from parsel.selector import Selector


class Response(object):
    def __init__(self, text: str, status: int, cookies: SimpleCookie):
        self.text = text
        self.status = status
        self.cookies = cookies
        self.json: dict = {}
        self.selector: Selector = None
        self.meta: dict = {}
